#!/bin/bash

# Set the value of ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS to 8
export ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS=8

# Initialize a counter for parallel jobs
parallel_jobs=0
max_jobs=12

# Iterate over image pairs
for i in {1..12}; do
  fixed="../Atlases/m$i.img"

  for j in {1..12}; do
    if [ $j != $i ]; then
      moving="../Atlases/m$j.img"
      output_prefix="outputs/deformed_${i}_${j}_seg.img"
      warp_prefix="outputs/output_${i}_${j}"

      greedy -d 3 -ri LABEL 0.2vox -rf $fixed -rm $moving $output_prefix -r "${warp_prefix}_warp.nii.gz" "${warp_prefix}.mat" &

      # Increment the parallel jobs counter
      ((parallel_jobs++))

      # Check if the maximum parallel jobs limit is reached
      if [ $parallel_jobs -eq $max_jobs ]; then
        # Wait for all background jobs to complete
        wait
        # Reset the parallel jobs counter
        parallel_jobs=0
      fi
    fi
  done
done

# Wait for any remaining background jobs to complete
wait

